/*+@@file@@----------------------------------------------------------------*//*!
 \file		Vector.c
 \par Description 
            
 \par  Status: 
            
 \par Project: 
            
 \date		Created  on Sun Aug  9 18:49:56 2015
 \date		Modified on Sun Aug  9 18:49:56 2015
 \author	
\*//*-@@file@@----------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <windows.h>
#include "vector.h"

#define INITALLOCBLOCK 4096

/*+@@fnc@@----------------------------------------------------------------*//*!
 \brief		VectorGetElement
 \date		Created  on Tue Sep  1 17:24:39 2015
 \date		Modified on Tue Sep  1 17:24:39 2015
\*//*-@@fnc@@----------------------------------------------------------------*/
void __forceinline *VectorGetElement(LPVECTORHEAD v, size_t nElem)
{
	char *p = (char *)v;

	return (void *)(p + sizeof(VECTORHEAD) + (nElem * v->cElemSize));
}

/*+@@fnc@@----------------------------------------------------------------*//*!
 \brief		_VectorResize
 \date		Created  on Fri Aug 28 20:48:01 2015
 \date		Modified on Fri Aug 28 20:48:01 2015
\*//*-@@fnc@@----------------------------------------------------------------*/
static BOOL _VectorResize(LPVECTORHEAD *p, size_t nElems, size_t cElemSize)
{
	LPVECTORHEAD v = NULL;

	if ((*p && !nElems) || (!*p && !cElemSize))
		return FALSE;

	size_t size = (((*p ? nElems : 1)*(*p ? (*p)->cElemSize : cElemSize)) + INITALLOCBLOCK - 1) & ~(INITALLOCBLOCK - 1);

	v = realloc(*p, size);

	if (!v)
		return FALSE;

	v->cSpaceAlloced = size;
	v->nAlloced      = (v->cSpaceAlloced - sizeof(VECTORHEAD)) / (*p ? (*p)->cElemSize : cElemSize);

	if (!*p)
	{
		v->nUsed         = 0;
		v->cElemSize     = cElemSize;
	}

	*p = v;

	return TRUE;
}

/*+@@fnc@@----------------------------------------------------------------*//*!
 \brief		VectorCreate
 \details	Creates a vector instance
 \date		Created  on Sun Aug  9 18:49:52 2015
 \date		Modified on Sun Aug 23 12:35:27 2015
\*//*-@@fnc@@----------------------------------------------------------------*/
LPVECTORHEAD VectorCreate(size_t cElemSize)
{
	LPVECTORHEAD p = NULL;

	_VectorResize(&p, 0, cElemSize);

	return p;
}

/*+@@fnc@@----------------------------------------------------------------*//*!
 \brief		VectorDelete
 \details	Deatroys a vector
 \date		Created  on Sun Aug  9 18:50:25 2015
 \date		Modified on Sun Aug 23 12:43:26 2015
\*//*-@@fnc@@----------------------------------------------------------------*/
BOOL VectorDelete(LPVECTORHEAD *p)
{
	if (!*p)
		return FALSE;

	LPVECTORHEAD v = *p;
	free(v);
	*p = NULL;
	return TRUE;
}

/*+@@fnc@@----------------------------------------------------------------*//*!
 \brief		VectorIsEmpty
 \details	Returns whether the vector is empty (i.e. whether its size is 
            0).
 \date		Created  on Wed Aug 19 12:43:40 2015
 \date		Modified on Sun Aug 23 12:36:05 2015
\*//*-@@fnc@@----------------------------------------------------------------*/
BOOL VectorIsEmpty(LPVECTORHEAD p)
{
	return !p->nUsed;
}

/*+@@fnc@@----------------------------------------------------------------*//*!
 \brief		VectorSize
 \details	Returns the number of elements in the vector.
            This is the number of actual objects held in the vector, which 
            is not necessarily equal to its storage capacity.
 \date		Created  on Sun Aug  9 18:50:03 2015
 \date		Modified on Sun Aug 23 12:36:46 2015
\*//*-@@fnc@@----------------------------------------------------------------*/
size_t VectorSize(LPVECTORHEAD p)
{
	return p->nUsed;
}

/*+@@fnc@@----------------------------------------------------------------*//*!
 \brief		VectorCapacity
 \details	Return size of allocated storage capacity.
            Returns the size of the storage space currently allocated for 
            the vector, expressed in terms of elements.
            This capacity is not necessarily equal to the vector size. It 
            can be equal or greater, with the extra space allowing to 
            accommodate for growth without the need to reallocate on each 
            insertion.
            Notice that this capacity does not suppose a limit on the size 
            of the vector. When this capacity is exhausted and more is 
            needed, it is automatically expanded by the container 
            (reallocating it storage space). The theoretical limit on the 
            size of a vector is given by member max_size.
            The capacity of a vector can be explicitly altered by calling 
            member VectorReserve.
 \date		Created  on Sun Aug 23 16:13:52 2015
 \date		Modified on Sun Aug 23 16:13:52 2015
\*//*-@@fnc@@----------------------------------------------------------------*/
size_t VectorCapacity(LPVECTORHEAD p)
{
	return p->nAlloced;
}

/*+@@fnc@@----------------------------------------------------------------*//*!
 \brief		VectorBack
 \details	Returns a reference to the last element in the vector.
            Unlike VectorEnd, which returns an iterator just past 
            this element, this function returns a direct reference.
            Calling this function on an empty container causes undefined 
            behavior.
 \date		Created  on Sun Aug  9 18:50:09 2015
 \date		Modified on Sun Aug 23 12:37:35 2015
\*//*-@@fnc@@----------------------------------------------------------------*/
void *VectorBack(LPVECTORHEAD p)
{
	return VectorGetElement(p, p->nUsed-1);
}

/*+@@fnc@@----------------------------------------------------------------*//*!
 \brief		VectorAt
 \details	Access element.
            Returns a reference to the element at position n in the 
            vector.
            The function automatically checks whether n is within the 
            bounds of valid elements in the vector, throwing an 
            out_of_range exception if it is not (i.e., if n is greater or 
            equal than its size). This is in contrast with member 
            operator[], that does not check against bounds.

 \date		Created  on Sun Aug 23 16:17:12 2015
 \date		Modified on Sun Aug 23 16:17:12 2015
\*//*-@@fnc@@----------------------------------------------------------------*/
void *VectorAt(LPVECTORHEAD p, size_t n)
{
	if (n >= p->nUsed)
		return NULL;
	return VectorGetElement(p, n);
}

/*+@@fnc@@----------------------------------------------------------------*//*!
 \brief		VectorData
 \details	Access data.
            Returns a direct pointer to the memory array used internally by 
            the vector to store its owned elements.
            Because elements in the vector are guaranteed to be stored in 
            contiguous storage locations in the same order as represented 
            by the vector, the pointer retrieved can be offset to access 
            any element in the array.
 \date		Created  on Sun Aug 23 16:20:37 2015
 \date		Modified on Sun Aug 23 16:20:37 2015
\*//*-@@fnc@@----------------------------------------------------------------*/
void *VectorData(LPVECTORHEAD p)
{
	if (!p->nUsed)
		return NULL;
	return VectorGetElement(p, 0);
}

/*+@@fnc@@----------------------------------------------------------------*//*!
 \brief		VectorPopBack
 \details	Delete last element
            Removes the last element in the vector, effectively reducing 
            the container size by one.
            This destroys the removed element.
 \date		Created  on Sun Aug 23 12:39:06 2015
 \date		Modified on Sun Aug 23 12:39:06 2015
\*//*-@@fnc@@----------------------------------------------------------------*/
BOOL VectorPopBack(LPVECTORHEAD p)
{
	p->nUsed--;

	return TRUE;
}

/*+@@fnc@@----------------------------------------------------------------*//*!
 \brief		VectorPush_Back
 \date		Created  on Sun Aug  9 18:50:17 2015
 \date		Modified on Sun Aug  9 18:50:17 2015
\*//*-@@fnc@@----------------------------------------------------------------*/
/*+@@fnc@@----------------------------------------------------------------*//*!
 \brief		VectorPushBack
 \details	Add element at the end
            Adds a new element at the end of the vector, after its current 
            last element. The content of val is copied (or moved) to the 
            new element.
            This effectively increases the container size by one, which 
            causes an automatic reallocation of the allocated storage space 
            if -and only if- the new vector size surpasses the current 
            vector capacity.
 \date		Created  on Sun Aug 23 12:40:00 2015
 \date		Modified on Sun Aug 23 12:40:00 2015
\*//*-@@fnc@@----------------------------------------------------------------*/
BOOL VectorPushBack(LPVECTORHEAD *p, void *pData)
{
	LPVECTORHEAD v = *p;

	if (v->nAlloced <= v->nUsed)
	{
		if (!_VectorResize(&v, v->nUsed + 1, 0))
			return FALSE;

		*p = v;
	}
	memcpy(VectorGetElement(v, v->nUsed), pData, v->cElemSize);
	v->nUsed++;
	return TRUE;
}

/*+@@fnc@@----------------------------------------------------------------*//*!
 \brief		VectorClear
 \details	Clear content
            Removes all elements from the vector (which are destroyed), 
            leaving the container with a size of 0.
            A reallocation is not guaranteed to happen, and the vector 
            capacity is not guaranteed to change due to calling this 
            function. A typical alternative that forces a reallocation is 
            to use swap: 
                   VectorSwap(x);   // clear x reallocating 

 \date		Created  on Sun Aug  9 18:50:21 2015
 \date		Modified on Sun Aug 23 12:41:05 2015
\*//*-@@fnc@@----------------------------------------------------------------*/
BOOL VectorClear(LPVECTORHEAD *p)
{
	if (!_VectorResize(p, 1, 0))
		return FALSE;

	(*p)->nUsed = 0;

	return TRUE;
}

/*+@@fnc@@----------------------------------------------------------------*//*!
 \brief		VectorErase
 \details	Erase elements.
            Removes from the vector either a single element (position) or a 
            range of elements ([first,last)).
            This effectively reduces the container size by the number of 
            elements removed, which are destroyed.
            Because vectors use an array as their underlying storage, 
            erasing elements in positions other than the vector end causes 
            the container to relocate all the elements after the segment 
            erased to their new positions. This is generally an inefficient 
            operation compared to the one performed for the same operation 
            by other kinds of sequence containers (such as list or 
            forward_list).
 \date		Created  on Sun Aug 23 16:35:10 2015
 \date		Modified on Sun Aug 23 16:35:10 2015
\*//*-@@fnc@@----------------------------------------------------------------*/
BOOL VectorErase(LPVECTORHEAD *p, size_t first, size_t last)
{
	LPVECTORHEAD v = *p;

	if (first >= v->nUsed || last >= v->nUsed)
		return FALSE;

	size_t range = abs(last - first) + 1;

	memmove(VectorGetElement(v, max(first, last)), VectorGetElement(v, min(first, last)), range);

	v->nUsed -= range;

	return TRUE;
}

/*+@@fnc@@----------------------------------------------------------------*//*!
 \brief		VectorInsert
 \details	The vector is extended by inserting new elements before the 
            element at the specified position, effectively increasing the 
            container size by the number of elements inserted.
            This causes an automatic reallocation of the allocated storage 
            space if -and only if- the new vector size surpasses the 
            current vector capacity.
            Because vectors use an array as their underlying storage, 
            inserting elements in positions other than the vector end 
            causes the container to relocate all the elements that were 
            after position to their new positions. This is generally an 
            inefficient operation compared to the one performed for the 
            same operation by other kinds of sequence containers (such as 
            list or forward_list).
 \date		Created  on Mon Aug 24 00:18:25 2015
 \date		Modified on Mon Aug 24 00:18:25 2015
\*//*-@@fnc@@----------------------------------------------------------------*/
BOOL VectorInsert(LPVECTORHEAD *p, size_t position, size_t n, void *pData)
{
	LPVECTORHEAD v = *p;

	if ( (v->nUsed + n) >= v->nAlloced )
		if (!_VectorResize(&v, v->nUsed + n + 1, 0))
			return FALSE;

	*p = v;

	memmove(VectorGetElement(v, position), VectorGetElement(v, position + n), n);

	memcpy (VectorGetElement(v, position), pData, n);

	return TRUE;
}
